/*
 *   PCM time-domain equalizer
 *
 *   Copyright (C) 2002  Felipe Rivera <liebremx at users sourceforge net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   $Id: iir.c,v 1.4 2002/08/31 21:44:50 felipe Exp $
 */

#include "eq.h"

/* IIR filter coefficients */
#include "iir_cf.h"

/* History for two filters */
static sXYData data_history[EQ_MAX_BANDS][EQ_CHANNELS] __attribute__((aligned));
static sXYData data_history2[EQ_MAX_BANDS][EQ_CHANNELS] __attribute__((aligned));

/* Coefficients */
static sIIRCoefficients *iir_cf;

/* Gain for each band
 * values should be between -0.2 and 1.0 */
extern float gain[EQ_MAX_BANDS][EQ_CHANNELS];
/* Volume gain 
 * values should be between 0.0 and 1.0 
 * Use the preamp from XMMS for now
 * */
extern float preamp[EQ_CHANNELS];

/* Init the filter */
void init_iir()
{
	/* Select the coefficients based on the number of bands */
	if (eqcfg.band_num == 15)
		iir_cf = iir_cf15;
	else if (eqcfg.band_num == 31)
		iir_cf = iir_cf31;
	else if (eqcfg.band_num == 10 && eqcfg.use_xmms_original_freqs)
		iir_cf = iir_cforiginal10;
	else
		iir_cf = iir_cf10;

	/* Zero the history arrays */
	bzero(data_history, sizeof(sXYData) * EQ_MAX_BANDS);
	bzero(data_history2, sizeof(sXYData) * EQ_MAX_BANDS);
}

/* 
 * The main stuff
 */
__inline__ int iir(gpointer * d, gint length)
{
	gint16 *data = (gint16 *) * d;
	/* Indexes for the history arrays
	 * These have to be kept between calls to this function
	 * hence they are static */
	static gint i = 0, j = 2, k = 1;	

	gint index, band, channel;
	gint tempgint;
	float out[EQ_CHANNELS], scaledpcm[EQ_CHANNELS];
	/**
	 * IIR filter equation is
	 * y[n] = 2 * (alpha*(x[n]-x[n-2]) + gamma*y[n-1] - beta*y[n-2])
	 *
	 * NOTE: The 2 factor was introduced in the coefficients to save
	 * 			a multiplication
	 *
	 * This algorithm cascades two filters to get nice filtering
	 * at the expense of extra CPU cycles
	 */
	for (index = 0; index < length/BYTES_PS; index+=2)
	{
		/* For each channel */
		for (channel = 0; channel < EQ_CHANNELS; channel++)
		{
			scaledpcm[channel] = data[index+channel]/32768.0;
			out[channel] = 0;
			/* For each band */
			for (band = 0; band < eqcfg.band_num; band++)
			{
				/* Store Xi(n) */
				data_history[band][channel].x[i] = scaledpcm[channel];
				/* Calculate and store Yi(n) */
				data_history[band][channel].y[i] = 
               		   (
               	/* 		= alpha * [x(n)-x(n-2)] */
				iir_cf[band].alpha * ( data_history[band][channel].x[i]
               			-  data_history[band][channel].x[k])
               	/* 		+ gamma * y(n-1) */
               			+ iir_cf[band].gamma * data_history[band][channel].y[j]
               	/* 		- beta * y(n-2) */
               			- iir_cf[band].beta * data_history[band][channel].y[k]
						);
				/* 
				 * If youre wondering why I dont multiply by 2.0 the coefficients instead
				 * of doing it here.... well.. then the filter becomes very unstable and
				 * stops working. 
				 * This is not true anymore.. I was doing something wrong, now it works fine */
				 /* Calculate the gain  */
				out[channel] +=  data_history[band][channel].y[i]*gain[band][channel];
			} /* For each band */

			/* Filter again the signal */
			for (band = 0; band < eqcfg.band_num; band++)
			{
				/* Store Xi(n) */
				data_history2[band][channel].x[i] = out[channel];
				/* Calculate and store Yi(n) */
				data_history2[band][channel].y[i] = 
               		   (
               	/* y(n) = alpha * [x(n)-x(n-2)] */
				iir_cf[band].alpha * (data_history2[band][channel].x[i]
               			-  data_history2[band][channel].x[k])
               	/* 	    + gamma * y(n-1) */
               			+ iir_cf[band].gamma * data_history2[band][channel].y[j]
               	/* 		- beta * y(n-2) */
               			- iir_cf[band].beta * data_history2[band][channel].y[k]
						);
				/* Calculate the gain */
				out[channel] +=  data_history2[band][channel].y[i]*gain[band][channel];
			} /* For each band */

			/* Volume stuff
			   Scale down original PCM sample and add it to the filters output */
			out[channel] += scaledpcm[channel] * 0.25;
			/* Volume gain */
			out[channel] *= preamp[channel];

			/* Adjust the fraction value back to a PCM sample
			   Scale back to a 16bit signed int */
			out[channel] *= 32768.0;

			/* Round and convert to integer */
			tempgint = rintf(out[channel]);
				
			/* Limit the output */
			if (tempgint < -32768)
				data[index+channel] = -32768;
			else if (tempgint > 32767)
				data[index+channel] = 32767;
			else 
				data[index+channel] = tempgint;
		} /* For each channel */
		i++; j++; k++;
		if (i == 3) i = 0;
		else if (j == 3) j = 0;
		else k = 0;
	}/* For each pair of samples */

	return length;
}
